#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2022 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
import asyncio
import itertools
from functools import partial
from typing import Any
from typing import AsyncIterator
from typing import Dict
from typing import Generic
from typing import Iterable
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

import more_itertools
from fastapi.encoders import jsonable_encoder
from httpx import AsyncClient
from httpx import HTTPStatusError
from structlog import get_logger
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_random_exponential
from tqdm import tqdm

logger = get_logger()


class ModelClientException(Exception):
    pass


ModelBase = TypeVar("ModelBase")


class ModelClientBase(Generic[ModelBase]):
    upload_http_method: str
    path_map: Dict[str, str]
    async_httpx_client_class: Type[AsyncClient]

    def __init__(self, *args: Any, chunk_size: int = 10, **kwargs: Any) -> None:
        """Base ModelClient.

        Args:
            *args: Positional arguments passed through to AuthenticatedAsyncHTTPXClient.
            chunk_size: Size of the chunks objects are split into before being uploaded
             in parallel.
            **kwargs: Keyword arguments passed through to AuthenticatedAsyncHTTPXClient.
        """
        super().__init__()
        self.async_httpx_client = self.async_httpx_client_class(*args, **kwargs)
        self.chunk_size = chunk_size

    def get_object_url(
        self, obj: ModelBase, *args: Any, **kwargs: Any
    ) -> str:  # pragma: no cover
        return self.path_map[type(obj).__name__]

    def get_object_json(
        self, obj: Union[ModelBase, Any], *args: Any, **kwargs: Any
    ) -> Any:
        return jsonable_encoder(obj)

    @retry(
        retry=retry_if_exception_type(HTTPStatusError),
        reraise=True,
        wait=wait_random_exponential(multiplier=2, max=30),
        stop=stop_after_attempt(3),
        after=lambda rs: logger.warning(f"Upload failed ({rs.attempt_number}/3)."),
    )
    async def upload_object(self, obj: ModelBase, *args: Any, **kwargs: Any) -> Any:
        response = await self.async_httpx_client.request(
            self.upload_http_method,
            self.get_object_url(obj, *args, **kwargs),
            json=self.get_object_json(obj, *args, **kwargs),
        )
        response_json = response.json()
        try:
            response.raise_for_status()
        except HTTPStatusError as error:
            if "description" in response_json:
                raise HTTPStatusError(
                    message=response_json["description"],
                    request=error.request,
                    response=error.response,
                ) from error
            raise error
        return response_json

    async def upload_lazy(
        self, objs: Iterable[ModelBase], *args: Any, **kwargs: Any
    ) -> AsyncIterator[Any]:
        objs = list(objs)  # len() is unfortunately needed for proper progress bar
        with tqdm(total=len(objs), unit="object") as progress_bar:
            for object_type, group in itertools.groupby(objs, key=type):
                progress_bar.set_description(object_type.__name__)
                for chunk in more_itertools.chunked(group, n=self.chunk_size):
                    tasks = map(partial(self.upload_object, *args, **kwargs), chunk)
                    for task in asyncio.as_completed(tasks):
                        yield await task
                        progress_bar.update()

    async def upload(
        self, objs: Iterable[ModelBase], *args: Any, **kwargs: Any
    ) -> List[Any]:
        return [x async for x in self.upload_lazy(objs, *args, **kwargs)]

    async def __aenter__(self):  # type: ignore
        # TODO: fix return type to Self when we get to a Python version >=3.11
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        await self.async_httpx_client.aclose()



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-few-public-methods,protected-access
"""This module contains the Generic AMQPSystem."""
from .abstract import AbstractAMQPSystem
from .abstract import AbstractPublishMixin
from .abstract import AbstractRouter


class Router(AbstractRouter):
    """Generic Router."""

    register = AbstractRouter._register


class PublishMixin(AbstractPublishMixin):
    """Generic PublishMixin."""

    publish_message = AbstractPublishMixin._publish_message


class AMQPSystem(AbstractAMQPSystem[Router], PublishMixin):
    """Generic AMQPSystem."""

    router_cls = Router



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Any
from typing import Coroutine
from typing import Dict
from typing import no_type_check
from typing import Optional
from typing import Type
from typing import Union

import httpx
from gql import Client as GQLClient
from gql.client import AsyncClientSession
from gql.client import SyncClientSession
from gql.transport import AsyncTransport
from graphql import DocumentNode
from pydantic import AnyHttpUrl

from ..auth import AuthenticatedAsyncHTTPXClient
from ..auth import AuthenticatedHTTPXClient
from ..auth import keycloak_token_endpoint
from ..graph.transport import AsyncHTTPXTransport
from ..graph.transport import HTTPXTransport


class GraphQLClient(GQLClient):
    transport: Union[HTTPXTransport, AsyncHTTPXTransport]

    def __init__(
        self,
        url: str,
        client_id: str,
        client_secret: str,
        auth_realm: str,
        auth_server: AnyHttpUrl,
        *args: Any,
        sync: bool = False,
        httpx_client_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        """
        GQL Client wrapper providing defaults and automatic authentication for OS2mo.
        If you need a client with a persistent session, for example from a FastAPI
        application, consider the ``PersistentGraphQLClient`` subclass.

        Args:
            url: URL of the GraphQL server endpoint.
            client_id: Keycloak client id used for authentication.
            client_secret: Keycloak client secret used for authentication.
            auth_realm: Keycloak auth realm used for authentication.
            auth_server: URL of the Keycloak server used for authentication.
            *args: Extra arguments passed to the superclass init method.
            sync: If true, this client is initialised with a synchronous transport.
            httpx_client_kwargs: Extra keyword arguments passed to the HTTPX client.
            **kwargs: Extra keyword arguments passed to the superclass init method.

        Example:
            Asynchronously::

                client = GraphQLClient(
                    url="http://os2mo.example.org/graphql",
                    client_id="AzureDiamond",
                    client_secret="hunter2",
                    auth_realm="mordor",
                    auth_server="https://keycloak.example.org:8081/auth",
                )
                async with client as session:
                    query = gql(
                        ""'
                        query MOQuery {
                          ...
                        }
                        ""'
                    )
                    result = await session.execute(query)
                    print(result)

            Or synchronously::

                with GraphQLClient(sync=True) as session:
                    result = session.execute(query)
        """
        transport_cls: Type[Union[HTTPXTransport, AsyncHTTPXTransport]]
        client_cls: Type[Union[httpx.Client, httpx.AsyncClient]]

        if sync:
            transport_cls = HTTPXTransport
            client_cls = AuthenticatedHTTPXClient
        else:
            transport_cls = AsyncHTTPXTransport
            client_cls = AuthenticatedAsyncHTTPXClient

        transport = transport_cls(
            url=url,
            client_cls=client_cls,
            client_args=dict(
                client_id=client_id,
                client_secret=client_secret,
                token_endpoint=keycloak_token_endpoint(
                    auth_server=auth_server,
                    auth_realm=auth_realm,
                ),
                **(httpx_client_kwargs or {}),
            ),
        )

        super().__init__(*args, transport=transport, **kwargs)  # type: ignore[misc]


class PersistentGraphQLClient(GraphQLClient):
    """
    GraphQLClient with persistent transport session. Since the session is shared, it is
    the responsibility of the caller to call/await ``close()``/``aclose()`` when done.

    Example:
        Example usage in a FastAPI application. The global client is created in a module
        called ``clients.py``::

            graphql_client = PersistentGraphQLClient(
                url=f"{settings.mo_url}/graphql",
                client_id=settings.client_id,
                client_secret=settings.client_secret,
                auth_realm=settings.auth_realm,
                auth_server=settings.auth_server,
            )


        Using the client from anywhere::

            @app.get("/test")
            async def test():
                return await graphql_client.execute(...)

        We must make sure to close the client on shutdown. FastAPI makes this very easy
        using a ``shutdown`` signal in ``app.py``::

            def create_app():
                app = FastAPI()

                @app.on_event("shutdown")
                async def close_clients():
                    await graphql_client.aclose()

                return app
    """

    @no_type_check
    def execute(
        self, document: DocumentNode, *args: Any, **kwargs: Any
    ) -> Union[Dict, Coroutine[None, None, Dict]]:
        """
        Execute the provided document AST against the remote server using the transport
        provided during init.

        Either the transport is sync, and we execute the query synchronously directly
        OR the transport is async, and we return an awaitable coroutine which executes
        the query. In any case, the caller can ``execute(...)`` or ``await execute()``
        as expected from the call context.

        Args:
            document: The GraphQL request query.
            *args: Extra arguments passed to the transport execute method.
            **kwargs: Extra keyword arguments passed to the transport execute method.

        Returns: Dictionary (or coroutine) containing the result of the query.
        """
        if isinstance(self.transport, AsyncTransport):
            return self.execute_async(document, *args, **kwargs)
        return self.execute_sync(document, *args, **kwargs)

    def __enter__(self) -> SyncClientSession:
        """Persist the GraphQL client session by only opening it once."""
        if not hasattr(self, "session"):
            super().__enter__()
        return self.session  # type: ignore

    async def __aenter__(self) -> AsyncClientSession:
        """Persist the GraphQL client session by only opening it once."""
        if not hasattr(self, "session"):
            await super().__aenter__()
        return self.session  # type: ignore

    def __exit__(self, *args: Any) -> None:
        """We're a persistent client: Don't close the transport when exiting."""
        pass

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        """We're a persistent client: Don't close the transport when exiting."""
        pass

    def close(self) -> None:
        """
        Close the transport session.
        """
        super().__exit__()

    async def aclose(self) -> None:
        """
        Close the async transport session.
        """
        await super().__aexit__(None, None, None)



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Dict
from typing import List
from typing import TypeVar

from .ensure_hashable import ensure_hashable


DictKeyType = TypeVar("DictKeyType")
DictValueType = TypeVar("DictValueType")


def transpose_dict(
    dicty: Dict[DictKeyType, DictValueType]
) -> Dict[DictValueType, List[DictKeyType]]:
    """Transpose a dictionary, such that keys become values and values become keys.

    *Note: Keys actually become a list of values, rather than plain values, as value
           uniqueness is not guaranteed, and thus multiple keys may have the same
           value.*

    Example:
        ```Python
        test_dict = {'test_key1': 'test_value1'}
        tdict = transpose_dict(test_dict)
        assert tdict == {"test_value1": ["test_key1"]}
        ```

    Example:
        ```Python
        test_dict = {
            "test_key1": "test_value1",
            "test_key2": "test_value2",
            "test_key3": "test_value1",
        }
        tdict = transpose_dict(test_dict)
        assert tdict == {
            "test_value1": ["test_key1", "test_key3"],
            "test_value2": ["test_key2"]
        }
        ```

    Args:
        dicty: Dictionary to be transposed.

    Returns:
        Transposed dictionary.

        *Note: Some values may be converted to hashable equivalentes using
               `ensure_hashable` to ensure they can be used as keys.*
    """

    # Reverse the dict
    reversed_dict: Dict[DictValueType, List[DictKeyType]] = dict()
    for key, value in dicty.items():
        # Ensure all new-key (value) is hashable
        new_key = ensure_hashable(value)
        reversed_dict[new_key] = reversed_dict.get(new_key, []) + [
            key,
        ]
    return reversed_dict



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from functools import wraps
from inspect import signature
from typing import Any
from typing import Callable
from typing import Tuple
from typing import TypeVar

CallableReturnType = TypeVar("CallableReturnType")


def has_self_arg(func: Callable) -> bool:
    """Return `True` if the given callable `func` needs `self` explicitly passed.

    Example:
        ```Python
        class Classy:
            def method(self):
                pass

            @classmethod
            def classmethod(cls):
                pass

            @staticmethod
            def staticmethod():
                pass

        def function():
            pass

        classy = Classy()
        # Can be passed implicitly by calling with 'classy.'
        assert has_self_arg(classy.method) is False
        assert has_self_arg(Classy.method) is True

        # These never need 'self' passed explicitly
        assert has_self_arg(classy.classmethod) is False
        assert has_self_arg(Classy.classmethod) is False
        assert has_self_arg(classy.staticmethod) is False
        assert has_self_arg(Classy.staticmethod) is False
        assert has_self_arg(function) is False
        ```

    Args:
        func (function): The function or method to check.

    Returns:
        Whether the provided function needs `self` passed explicitly.
    """
    args = list(signature(func).parameters)
    return bool(args) and args[0] in ("self")


def apply(func: Callable[..., CallableReturnType]) -> Callable[..., CallableReturnType]:
    """Function decorator to apply tuple to function.

    Example:
        ```Python
        @apply
        def dual(key, value):
            return value

        print(dual(('k', 'v')))  # --> 'v'
        ```

    Args:
        func: The function to apply arguments for.

    Returns:
        The function which has had it argument applied.
    """

    if has_self_arg(func):

        @wraps(func)
        def wrapper(self: Any, tup: Tuple) -> CallableReturnType:
            return func(self, *tup)

    else:

        @wraps(func)
        def wrapper(tup: Tuple) -> CallableReturnType:
            return func(*tup)

    return wrapper



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from functools import partial
from functools import wraps
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from jinja2 import Template


def requires_jinja(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Optional[Any]) -> Any:
        return func(*args, **kwargs)

    return wrapper


InnerFilterType = Callable[[List[str], List[Any]], bool]
FilterType = Callable[[List[Any]], bool]


def string_to_bool(v: str) -> bool:
    """Convert a string to a boolean.

    Returns `True` if lowercase version of the string is one of:
    `"yes"`, `"true"`, `"1"` or `"1.0"`

    *Note: The values for truth were chosen semi-arbitrarily.*

    Args:
        v: The string value to be converted to a boolean.

    Returns:
        The boolean interpretation of the string.
    """
    return v.lower() in ("yes", "true", "1", "1.0")


@requires_jinja
def jinja_filter(
    template: Template, tuple_keys: List[str], tuple_values: List[Any]
) -> bool:
    """Filter function to evaluate the filter on the provided argument list.

    *Note: Utilizes `tuple_keys` as keys to map the incoming list of arguments to
           jinja2 key-value context variables.*

    *Note: Utilizes `string_to_bool` to convert the output of the template into a
           boolean value.*

    Args:
        template: The `jinja2` template to put our context into.
        tuple_keys: List of keys to put into the `jinja2` template context.
        tuple_values: List of values to put into the `jinja2` template context.

    Returns:
        Whether the filter passed or not.
    """
    context: Iterator[Tuple[str, Any]] = zip(tuple_keys, tuple_values)
    context_dict: Dict[str, Any] = dict(context)
    result: str = template.render(**context_dict)
    return string_to_bool(result)


@requires_jinja
def create_filter(
    jinja_string: str,
    tuple_keys: List[str],
) -> FilterType:
    """Convert a `jinja2` filter strings into a filter function.

    Args:
        jinja_string: The filter string to be converted into a function.
        tuple_keys: List of keys to put into the `jinja2` template context.

    Returns:
        The generated filter function, to be called with `tuple_values`.
    """
    filter_function: FilterType = partial(
        jinja_filter, Template(jinja_string), tuple_keys
    )
    return filter_function


@requires_jinja
def create_filters(
    jinja_strings: List[str],
    tuple_keys: List[str],
) -> List[FilterType]:
    """Convert a list of `jinja2` filter strings into filter functions.

    For more details see `create_filter`.

    Args:
        jinja_strings: A list of filter strings to be converted into functions.
        tuple_keys: List of keys to put into the `jinja2` template contexts.

    Returns:
        A list of generated filter function, to be called with `tuple_values`.
    """

    filter_functions: Iterator[FilterType] = map(
        partial(create_filter, tuple_keys=tuple_keys), jinja_strings
    )
    return list(filter_functions)



# SPDX-FileCopyrightText: 2022 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import logging
import sys
from enum import auto
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import structlog
from pydantic import BaseSettings
from pydantic import Extra
from pydantic.env_settings import SettingsSourceCallable
from structlog.processors import CallsiteParameter

from .load_settings import load_settings

logger = logging.getLogger(__name__)


def _get_json_settings_source(prefix: str) -> SettingsSourceCallable:
    """Create a Pydantic settings source which reads the DIPEX `settings.json`

    Args:
        prefix (str): Retrieve only the settings in `settings.json` whose key match this
                      prefix.
    """

    def settings_source(base_settings: BaseSettings) -> Dict[str, Any]:
        # The actual settings source callable

        try:
            all_settings = load_settings()
        except FileNotFoundError:
            # `print()` is used instead of `logger.warning()` here, as logging is
            # probably not yet configured at this point.
            print("Could not load 'settings.json', using settings from environment")
            return {}

        # Retrieve all settings matching `prefix`, and convert dots in setting keys to
        # underscores, so they become valid setting names for Pydantic settings.
        settings: Dict[str, Any] = {
            key.replace(".", "_"): val
            for key, val in all_settings.items()
            if key.startswith(prefix)
        }

        # Add log level to settings, if defined
        if "log_level" in all_settings:
            settings["log_level"] = all_settings["log_level"]

        return settings

    return settings_source


class LogLevel(Enum):
    """Represent log levels of the Python `logging` library as an `Enum`.

    This allows us to read the desired log level in `IntegrationSettings`  without
    additional parsing or conversion.
    """

    def _generate_next_value_(name, start, count, last_values):  # type: ignore
        # Generate enum elements whose key and values are identical.
        # See: https://docs.python.org/3.9/library/enum.html#using-automatic-values
        # Must be defined *before* the enum elements to work correctly.
        return name

    NOTSET = auto()
    DEBUG = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()
    CRITICAL = auto()


def _dont_log_graphql_responses(_: Any, __: str, event_dict: dict) -> dict:
    """Drop logs from `BaseHTTPXTransport._decode_response` (in
    `raclients.graph.transport`), which logs *all* GraphQL responses at DEBUG level.
    (https://git.magenta.dk/rammearkitektur/ra-clients/-/blob/master/raclients/graph/transport.py#L117)
    """
    module: str | None = event_dict.get("module")
    func_name: str | None = event_dict.get("func_name")
    if module == "transport" and func_name == "_decode_response":  # pragma: no cover
        raise structlog.DropEvent
    return event_dict


class JobSettings(BaseSettings):
    """Base class for defining the settings of a given OS2MO integration job.

    Each integration should define its settings like this:
    >>> class SqlExportSettings(JobSettings):
    >>>
    >>>     class Config:
    >>>         # Optional: Only use settings from settings.json if they match this
    >>>         # prefix.
    >>>         settings_json_prefix = "exporters.actual_state"

    And then read its settings like this:
    >>> settings = SqlExportSettings()

    And configure logging according to the settings like this:
    >>> settings.start_logging_based_on_settings()
    """

    mora_base = "http://mo:5000"
    client_id: str = "dipex"
    client_secret: Optional[str] = None
    auth_realm: str = "mo"
    auth_server: str = "http://keycloak:8080/auth"

    log_level: LogLevel = LogLevel.DEBUG
    log_format: str = (
        "%(levelname)s %(asctime)s %(filename)s:%(lineno)d:%(name)s: %(message)s"
    )

    sentry_dsn: Optional[str] = None

    class Config:
        # Configuration attributes defined by the Pydantic `Config` class
        extra: Extra = Extra.allow
        env_file_encoding: str = "utf-8"
        use_enum_values: bool = True

        # Additional configuration attributes defined by us.
        settings_json_prefix: str = ""

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            """Add settings source which reads settings from 'settings.json'"""
            json_settings = _get_json_settings_source(cls.settings_json_prefix)
            return (
                init_settings,
                env_settings,
                json_settings,
                file_secret_settings,
            )

    def start_logging_based_on_settings(self) -> None:
        """Configure Python `logging` library as well as `structlog` logging according
        to the specified log level."""
        self._configure_python_logging()
        self._configure_structlog_logging()

    def _configure_python_logging(self) -> None:
        # Based on https://stackoverflow.com/a/14058475

        # Get root logger and set its log level
        root: logging.Logger = logging.getLogger()
        root.setLevel(self._get_log_level_numeric_value())

        # Create handler logging to stdout, and set its log level
        handler: logging.StreamHandler = logging.StreamHandler(sys.stdout)
        handler.setLevel(self._get_log_level_numeric_value())

        # Set the log format of the handler
        formatter: logging.Formatter = logging.Formatter(self.log_format)
        handler.setFormatter(formatter)

        # Add the handler logging to stdout to the root logger
        root.addHandler(handler)

    def _configure_structlog_logging(self) -> None:
        # Based on: https://www.structlog.org/en/stable/logging-best-practices.html

        shared_processors: List[Any] = [
            structlog.processors.CallsiteParameterAdder(
                [CallsiteParameter.MODULE, CallsiteParameter.FUNC_NAME],
            ),
            _dont_log_graphql_responses,
        ]
        if sys.stderr.isatty():
            # Pretty printing when we run in a terminal session.
            # Automatically prints pretty tracebacks when "rich" is installed
            processors = shared_processors + [structlog.dev.ConsoleRenderer()]
        else:  # pragma: no cover
            # Print JSON when we run, e.g., in a Docker container.
            # Also print structured tracebacks.
            processors = shared_processors + [structlog.processors.JSONRenderer()]

        structlog.configure(
            processors,
            # Only log `structlog` output at the configured log level, or higher
            wrapper_class=structlog.make_filtering_bound_logger(
                self._get_log_level_numeric_value()
            ),
        )

    def _get_log_level_numeric_value(self) -> int:
        reverse_mapping = {value: key for key, value in logging._levelToName.items()}
        return reverse_mapping[str(self.log_level)]



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import re
from typing import cast
from typing import Dict
from typing import Iterator
from typing import Pattern


def multiple_replace_compile(replacement_dict: Dict[str, str]) -> Pattern:
    """Make a regex pattern for finding all keys in replacement dict.

    Calling this directly with `multiple_replace_run` allows one to generate the regex
    only once, but using it multiple times, which is advantageous for performance.

    Args:
        replacement_dict: Dictionary of replacements. Keys are made into a regex.

    Returns:
        Regex matching all keys in replacement dict.
    """
    # Replacing empty string is a mess
    keys = replacement_dict.keys()
    if "" in keys:
        raise ValueError("Cannot replace empty string")

    # Make a regex pattern, which matches all (escaped) keys
    escaped_keys = map(re.escape, keys)  # type: ignore
    pattern = re.compile("|".join(cast(Iterator[str], escaped_keys)))

    return pattern


def multiple_replace_run(
    pattern: Pattern, replacement_dict: Dict[str, str], string: str
) -> str:
    """Run a a regex pattern to replace matches.

    Calling this directly with a regex from `multiple_replace_compile` allows one to
    only generate the regex once, but using it multiple times, which is advantageous
    for performance.

    Args:
        pattern: A regex pattern produced by `multiple_replace_compile`, using the
            same `replacment_dict` provided here.
        replacement_dict: Dictionary of replacements.
            Keys are replaced with their corresponding values.
        string: The string to make replacements in.

    Returns:
        Modified string, with all replacements made.
    """
    # For each match, replace found key with corresponding value
    return cast(
        str, pattern.sub(lambda x: replacement_dict.get(x.group(0), ""), string)
    )


def multiple_replace(replacement_dict: Dict[str, str], string: str) -> str:
    """Make multiple replacements in string.

    *Note: Glues together `multiple_replace_compile` and `multiple_replace_run`.*

    Example:
        ```Python
        result = multiple_replace({"like": "love", "tea": "coffee"}, "I like tea")
        assert result == "I love coffee"

        result = multiple_replace(
            {"I": "love", "love": "eating", "eating": "spam"},
            "I love eating"
        )
        assert result == "love eating spam"
        assert result != "spam spam spam"
        ```

    Args:
        replacement_dict: Dictionary of replacements.
            Keys are replaced with their corresponding values.
        string: The string to make replacements in.

    Returns:
        Modified string, with all replacements made.
    """
    pattern = multiple_replace_compile(replacement_dict)
    return multiple_replace_run(pattern, replacement_dict, string)



#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2022 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Protocol
from typing import Union
from uuid import UUID

from fastapi.encoders import jsonable_encoder
from pydantic import AnyHttpUrl

from ..auth import AuthenticatedAsyncHTTPXClient
from ..auth import keycloak_token_endpoint
from ..modelclient.base import ModelClientBase


class MOBase(Protocol):
    def dict(self) -> dict:  # pragma: no cover
        ...

    @property
    def uuid(self) -> UUID:  # pragma: no cover
        ...


class ModelClient(ModelClientBase[MOBase]):
    upload_http_method = "POST"
    create_path_map: Dict[str, str] = {
        "Address": "/service/details/create",
        "Association": "/service/details/create",
        "Employee": "/service/e/create",
        "Engagement": "/service/details/create",
        "EngagementAssociation": "/service/details/create",
        "ClassWrite": "/service/f/{facet_uuid}/",
        "ITUser": "/service/details/create",
        "KLE": "/service/details/create",
        "Leave": "/service/details/create",
        "Manager": "/service/details/create",
        "OrganisationUnit": "/service/ou/create",
        "Role": "/service/details/create",
    }
    edit_path_map: Dict[str, str] = {
        "Address": "/service/details/edit",
        "Association": "/service/details/edit",
        "Employee": "/service/details/edit",
        "Engagement": "/service/details/edit",
        "ClassWrite": "/service/f/{facet_uuid}/",
        "ITUser": "/service/details/edit",
        "KLE": "/service/details/edit",
        "Leave": "/service/details/edit",
        "Manager": "/service/details/edit",
        "OrganisationUnit": "/service/details/edit",
        "Role": "/service/details/edit",
    }
    async_httpx_client_class = AuthenticatedAsyncHTTPXClient

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        auth_realm: str,
        auth_server: AnyHttpUrl,
        force: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        """MO ModelClient.

        Args:
            client_id: Keycloak client id used for authentication.
            client_secret: Keycloak client secret used for authentication.
            auth_realm: Keycloak auth realm used for authentication.
            auth_server: URL of the Keycloak server used for authentication.
            force: Bypass MO API validation.
            *args: Positional arguments passed through to ModelClientBase.
            **kwargs: Keyword arguments passed through to ModelClientBase.

        Example usage::

            async with ModelClient(
                base_url="http://mo:5000",
                client_id="AzureDiamond",
                client_secret="hunter2",
                auth_server=parse_obj_as(AnyHttpUrl,"https://keycloak.example.org/auth"),
                auth_realm="mordor",
            ) as client:
                r = await client.upload(objects)
        """
        super().__init__(
            *args,
            client_id=client_id,
            client_secret=client_secret,
            token_endpoint=keycloak_token_endpoint(
                auth_server=auth_server,
                auth_realm=auth_realm,
            ),
            **kwargs,
        )
        self.force = force

    def get_object_url(
        self, obj: MOBase, *args: Any, edit: bool = False, **kwargs: Any
    ) -> str:
        # Note that we additionally format the object's fields onto the path mapping to
        # support schemes such as /service/f/{facet_uuid}/, where facet_uuid is
        # retrieved from obj.facet_uuid.
        path_map = self.edit_path_map if edit else self.create_path_map
        path = path_map[type(obj).__name__].format_map(obj.dict())
        return f"{path}?force={int(self.force)}"

    def get_object_json(
        self, obj: Union[MOBase, Any], *args: Any, edit: bool = False, **kwargs: Any
    ) -> Any:
        # 'jsonable_encoder' is used directly on the obj, as 'exclude_defaults' doesn't
        # work when the object is nested within a dict.
        # TODO: Ideally we'd want to use 'exclude_unset' instead of 'exclude_defaults',
        #  but it doesn't seem to work properly with our models.
        if edit:
            obj = {
                "uuid": obj.uuid,
                "type": obj.type_,  # type: ignore[union-attr]
                "data": jsonable_encoder(obj, exclude_defaults=True),
            }
        return super().get_object_json(obj, *args, **kwargs)

    async def edit(
        self, objs: Iterable[MOBase], *args: Any, **kwargs: Any
    ) -> List[Any]:
        return await self.upload(objs, *args, edit=True, **kwargs)
